//
// DIVERSE graphics interface -- Performer ---- version 2.0.5
// released: Sat May 18 10:54:26 EDT 2002
// The DIVERSE HOMPAGE is at: www.diverse.vt.edu
// 
// Copyright (C) 2000, 2001, 2002  John Kelso, Lance Arsenault and Ron Kriz
// 
// This software, DIVERSE graphics interface -- Performer (DPF),
// is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License (LGPL) as
// published by the Free Software Foundation; either version 2.1 of
// the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this software, in the top level source directory in
// a file named "COPYING"; if not, see it at:
// http://www.gnu.org/copyleft/lesser.html or write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
//  Modified version of collision detection to "snap" to the terrain
//  from any position.  
//
//  This DSO adds collision detection and terrain following to the currently
//  loaded navigation DSO. The 'T' key toggles terrain following, and the 'M'
//  key toggles collision detection (M for solid Matter).  Collision detection
//  will keep the navigation position from passing through object in the scene.
//  Terrain following will make the user stick to the floor
//
//  By default, collision detection and terrain following are both enabled when
//  the DSO is loaded.  If the TERRAIN symbol is defined at compile time, only
//  terrain following is enabled.  Collision detection can still be toggled in
//  this case.
//
//  This DSO works best when a navigation DSO such as desktopFlyNav or
//  wandJoystickNav is loaded.  I would not recommed using the
//  collisionDetectin DSO when a navigation such as trackballNav is loaded.

#include <stdlib.h>
#include <math.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>

#define TERRAIN

#ifndef TERRAIN
#  define NAVNAME CollisionDetection
#  define NAVNAMESTRING "CollisionDetection"
#endif

#ifdef TERRAIN
#  define NAVNAME SnapTerrain
#  define NAVNAMESTRING "SnapTerrain"
#endif

//  Key to toggle collision detection
#define COLLISION_KEY XK_M

//  Key to toggle terrain following
#define TERRAIN_KEY XK_T

class NAVNAME : public dtkAugment {

public:
  NAVNAME(dpf *app_in) ;
  ~NAVNAME(void) ;
  int postConfig(void);
  int postFrame(void);

private:
  dpf* app;
  dpfDisplay *display;

  //  Matrices to convert between a navigation location and a position
  //  in world coordinates
  pfMatrix pivotMat, invPivotMat;
  
  //  CollisionNav object that does that actual collision detection
  dpfCollisionNav* collision;

  //  Key code to toggle terrain following
  KeyCode terrain_key;

  //  Key code to toggle collision detection
  KeyCode collision_key;

  //  Keyboard to read from
  dtkInXKeyboard *keyboard;

  //  Dequeuer for the keyboard
  dtkDequeuer *dq;
} ;

NAVNAME::NAVNAME(dpf *app_in) :
  //  The name is set the same for both CollisionDetection and TerrainOnly.
  //  We don't want them both loaded at the same time.
  dtkAugment("collisionDetection")
{
  app = app_in;
  display = app->display();
  setDescription("collision detection and terrain following DSO") ;
  validate();
} ;

NAVNAME::~NAVNAME(void) {

  if (collision) {
    delete collision;
    collision = NULL;
  }

  if (dq) {
    delete dq;
    dq = NULL;
  }

}

int NAVNAME::postConfig(void) {
  collision = NULL;
  keyboard = NULL;
  dq = NULL;
 
  //  Find where (0,0,0) in dpf units is in Ether coordinates
  pfCoord c0;
  c0.xyz.set(0.0f, 0.0f, 0.0f);
  c0.hpr.set(0.0f, 0.0f, 0.0f);
  
  display->dpfToEther(&c0);
  pivotMat.makeCoord(&c0);
  invPivotMat.invertAff(pivotMat);

  //  Initialize the CollisionNav object
  collision = new dpfCollisionNav(app);

#ifndef TERRAIN
  collision->enableCollision();
#endif

  collision->enableTerrain();

  collision->setLocation(c0.xyz);

  //  Set a really high falling speed
  collision->setFallSpeed(50000000);

  //  Get the keyboard
  keyboard = (dtkInXKeyboard *) app->get("xkeyboard", DTKINXKEYBOARD_TYPE);
  if(!keyboard) 
    return DTK_CALLBACK_ERROR;
  keyboard->queue();

  //  Get the X key codes for toggling
  if(!(collision_key = keyboard->getKeyCode(COLLISION_KEY)))
    {
      dtkMsg.add(DTKMSG_ERROR,"CollisionDetection::postConfig() failed" 
                              " to get x key code for COLLISION_KEY.\n");
      return DTK_CALLBACK_ERROR;
    }

  if(!(terrain_key = keyboard->getKeyCode(TERRAIN_KEY)))
    {
      dtkMsg.add(DTKMSG_ERROR,"CollisionDetection::postConfig() failed"
                              " to get x key code for TERRAIN_KEY.\n");
      return DTK_CALLBACK_ERROR;
    }
  
  //  Start the dequeuer for the keyboard
  dq = new dtkDequeuer(app->record());
  if(!dq || dq->isInvalid()) 
    return  DTK_CALLBACK_ERROR;

  return DTK_CONTINUE ; 
}


int NAVNAME::postFrame(void) {

  //  Check the keyboard for toggle keys
  dtkRecord_event *event;
  KeyCode keyCode;

  while ((event = dq->getNextEvent(keyboard))) {
    int pressed = keyboard->read(&keyCode, event);

    if (pressed && keyCode == collision_key) {

      //  Toggle collision detection
      if (collision->getCollision()) {
        collision->disableCollision();
	dtkMsg.add(DTKMSG_INFO, NAVNAMESTRING"::postFrame():  "
		   "Collision detection disabled\n") ;
      }
      else {
        collision->enableCollision();
	dtkMsg.add(DTKMSG_INFO, NAVNAMESTRING"::postFrame():  "
		   "Collision detection enabled\n") ;
      }

    }
    else if (pressed && keyCode == terrain_key) {

      //  Toggle terrain following
      if (collision->getTerrain()) {
        collision->disableTerrain();
	dtkMsg.add(DTKMSG_INFO, NAVNAMESTRING"::postFrame():  "
		   "Terrain following disabled\n") ;
      }
      else {
        collision->enableTerrain();
	dtkMsg.add(DTKMSG_INFO, NAVNAMESTRING"::postFrame():  "
		   "Terrain following enabled\n") ;
      }

    }

  }

  dpfNav* currentNav = display->currentNav;

  //printf ("snapTerrain: the current frame  =%d\n", pfGetFrameCount());
  if (currentNav) {

    //  Find the current navigation point in world coordinates
    pfCoord worldPoint;
    worldPoint.xyz.set(0.0f, 0.0f, 0.0f);
    worldPoint.hpr.set(0.0f, 0.0f, 0.0f);

    display->dpfToWorld(&worldPoint);

    //  Check this point for collision detection
    //  Jump above the terrain, then intersect below us
    if (collision->getTerrain()) {
      worldPoint.xyz[2] = 4000;
      collision->setLocation(worldPoint.xyz);
    }
    collision->tryMoveLocation(worldPoint.xyz);

    worldPoint.xyz = collision->getNewLocation();
    
    pfMatrix locMat;
    locMat.makeCoord(&worldPoint);
    locMat = invPivotMat * locMat;
    locMat.getOrthoCoord(&(currentNav->location));
  }
  return DTK_CONTINUE ; 
}


/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new NAVNAME(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}


