#include "fogDso.h"
#include <FL/Fl_Value_Slider.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Color_Chooser.h>
#include <Performer/pr/pfFog.h>

#define DESCRIPTION "Allows users to modify the visibility of the scene," \
		    "by changing the fog values\n"
#define R 0
#define G 1
#define B 2

static float r = .772, g = .807, b = 1.0;
static Fl_Window* make_window ();

//  Variables allocated in performer shared memory so they can be seen by
//  the draw process
static float* fogValue = NULL;
static pfVec3* fogColor = NULL;
static bool* isOn = NULL;
static bool* changed = NULL;  // has fog changed since last frame?

/********************************
 *	  Class Methods
 *******************************/

// constructor 
fogDso::fogDso(dpf *app_in):dtkAugment("fog") {
  setDescription(DESCRIPTION);
  app = app_in;
  fog = NULL;
  validate();
}

// destructor
fogDso::~fogDso() {
  if(guiWin) {
    delete guiWin;
    guiWin = NULL;
  }
}

//  Draw callback to set the fog distance and color.  Fog values may only
//  be changed from the draw process
int preDraw(pfTraverser* trav, void* data) {
  static bool firstFrame= true;

  //  Set up the fog on the first preDraw call
  if (firstFrame) {

    pfEnable(PFEN_FOG);
    
    pfFog* fog = new pfFog();
    fog->setFogType(PFFOG_VTX_EXP2);

    fog->apply();

    pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_ON);

    firstFrame= false;
  }
  

  //  Set the earth sky to be the same color as fog
  pfChannel* channel = (pfChannel*) data;
  pfEarthSky* eSky = channel->getESky();
  if (eSky) {
    eSky->setMode(PFES_BUFFER_CLEAR, PFES_FAST);
    eSky->setColor(PFES_CLEAR, (*fogColor)[0], (*fogColor)[1], (*fogColor)[2], 1);
  }
  
  //  Apply the current fog settings
  pfFog *fog = pfGetCurFog();

  if(fog) {

    pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_OFF);
    
    if(*isOn){
  
      //  Turn on fog, set color and range
      pfEnable(PFEN_FOG);
      fog->setRange(0.0f,*fogValue);
      fog->setColor((*fogColor)[0], (*fogColor)[1], (*fogColor)[2]);
      fog->apply();

    }
    else {
      //  Turn off fog
      pfDisable(PFEN_FOG);
    }

    pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_ON);
  }
  else {
    printf ("ERROR: pfGetCurFog() returned null\n");
  }

  return PFTRAV_CONT;
}

int fogDso::preConfig() {
  //  Set up variables in performer shared memory
  void* arena = pfGetSharedArena();

  fogValue = (float*) pfMalloc (sizeof(float), arena);
  *fogValue = 10000;

  fogColor = (pfVec3*) pfMalloc (sizeof(pfVec3), arena);
  fogColor->set(.772, .807, 1.0f);

  isOn = (bool*) pfMalloc(sizeof(bool), arena);
  *isOn = true;

  changed = (bool*) pfMalloc(sizeof(bool), arena);
  *changed = true;

  return DTK_REMOVE_CALLBACK;
}

// initialize everything
int fogDso::postConfig(){

  //  Data to the callback is a pointer to the first channel
  app->display()->world()->setTravData(PFTRAV_DRAW, 
      app->display()->first()->first()->first()->pfchan(0));
  app->display()->world()->setTravFuncs(PFTRAV_DRAW, preDraw, NULL);

  // create the gui
  guiWin = make_window();
  if(!guiWin) return DTK_CALLBACK_ERROR;

  return DTK_REMOVE_CALLBACK;
}

// check gui to see of stuff needs 
// to get done each frame
int fogDso::postFrame() {
  Fl::check();
  return DTK_CONTINUE;
}

/**********************************
 *	    Callbacks
 *********************************/

// turns the fog on and off
static void onOff_callback(Fl_Button* onOff, Fl_Value_Slider* s) {
  *isOn = !*isOn;
  *changed = true;

  if(*isOn){
    onOff->label("Fog Off");
  }
  else 
    onOff->label("Fog On");

}

// modifies the opacity of the fog with sliders
static void modifyVisibility_callback(Fl_Value_Slider* s, void*) {
 	
  *fogValue = s->value();
  *changed = true;
}

// changes the color of the fog according to rgb values
// from the Fl_Color_Chooser widget
static void modifyColor_callback(Fl_Color_Chooser *color, void*) {

  if(*isOn) {
    *changed = true;
  }  

  fogColor->set(color->r(), color->g(), color->b());
  
}

// reset the color back to the initial rgb values
static void reset_callback(Fl_Button *, void* color) {

  if(*isOn) {
    *changed = true;
  }

  fogColor->set(.772, .807, 1.0);
  ((Fl_Color_Chooser*)color)->rgb(.772,.807,1.0);
}

// hides the gui
static void hide_callback(Fl_Widget*, void* w) {
   ((Fl_Window*)w)->iconize();
}

// creates the gui
static Fl_Window* make_window() {
  Fl_Window* w;
  Fl_Color_Chooser *color;
  Fl_Value_Slider* s = NULL;
  { Fl_Window* o = new Fl_Window(300, 230, "Fog Value");
    w = o;
    { Fl_Value_Slider* o = s = new Fl_Value_Slider(5, 25, 290, 25, "Visibility:  ");
      o->type(5);
      o->selection_color(34);
      o->labeltype(FL_ENGRAVED_LABEL);
      o->minimum(50);
      o->step(1);
      o->maximum(100000);
      o->value(10000);
      o->align(FL_ALIGN_TOP_LEFT);
      o->callback((Fl_Callback*)modifyVisibility_callback);
    }
    { Fl_Color_Chooser *o = color = new Fl_Color_Chooser(5,75,290,115, "Fog Colors");
      o->rgb(r,g,b);
      o->labeltype(FL_ENGRAVED_LABEL);
      o->align(FL_ALIGN_TOP_LEFT);
      o->callback((Fl_Callback*)modifyColor_callback);
      o->when(FL_WHEN_CHANGED);     
    } 
    { Fl_Button* o = new Fl_Button(5, 195, 93, 30, "Fog Off");
      o->labeltype(FL_ENGRAVED_LABEL);  
      o->callback((Fl_Callback*)onOff_callback, (void*)s);
    }
    { Fl_Button* o = new Fl_Button(103, 195, 93, 30, "Reset");	//5, 195, 100, 30
      o->labeltype(FL_ENGRAVED_LABEL);
      o->callback((Fl_Callback*)reset_callback, (void*)color);
    } 
    { Fl_Button* o = new Fl_Button(201, 195, 93, 30, "Hide");
      o->labeltype(FL_ENGRAVED_LABEL);
      o->callback((Fl_Callback*)hide_callback, (void*) w);
    }
    o->end();
  }

  w->show();

  return w;
}

/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new fogDso(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}


