#include <stdio.h>
#include <dpf.h>
#include <dtk.h>
#include <dtk/dtkDSO_loader.h>
#include <string>
#include <vector>
using namespace std;
#include "Communicator.h"
// example generic dtkAugment DSO modified to be a file loader for 
// any set world
// Notice how simple it is to take an example program and modify it
struct Item
{
	string Name;                     //The description of the item
	pfDCS* Master;                   //Node that is added to the world
	pfDCS* Parent;                   //Node that holds the item
	pfDCS* Children;                 //Node that handles all the child pointers
	vector<Item> ChildList;          //List of children
};

class ItemDSO : public dtkAugment {
public:

  ItemDSO(dpf *app) ;
  ~ItemDSO(void) ;
  //During the post frame step handle everything
  int postFrame(void);

private:

  //Add an item into the world
  void AddItem(Message* AMessage);
  //Remove an item to the world
  void RemoveItem(const string& Name);
  //Handles the actual loading of an item in the world
  void LoadItem(Item& AnItem);
  //Handles the actual un-loading of an item in the world
  void UnLoadItem(Message* AMessage);
  //Application has been granted control of an item
  void Grant(Message* AMessage);
  //Revoke control of an object
  void Revoke(Message* AMessage);
  //Load items into the world (at startup)
  void LoadItems(Message* AMessage);
  //Load a specific item located at a url into the world
  void LoadURL(Message* AMessage);
  //Move an object around
  void MoveObject(Message* AMessage);
  //Has the item that I have control over moved?
  bool ItemMoved();
  //Write the position of my controlled item to the world
  void WritePosition();
  //Associate items
  void AssociateItems(Message* AMessage);
  void DisassociateItems(Message* AMessage);
  //Setup all the pointers for the item
  void ConstructItem(Item& New);

  //Quick pointer to the item I control
  Item ControlledItem;

  //The application that we use
  dpf* App;
  //The communications object
  Communicator* Data;
  //The hostname of the machine
  string HName;
  //The list of items
  vector<Item> Items;
  //The file path for loading objects
  string FilePath;
  //Am I controlling something
  bool Control;
};

void ItemDSO::ConstructItem(Item& New)
{
	//Create nodes
	New.Master   = new pfDCS;
	New.Parent   = new pfDCS;
	New.Children = new pfDCS;
	//Put nodes in proper relationship
	New.Master->addChild(New.Parent);
	New.Master->addChild(New.Children);
	//Display items in the world
	App->display()->world()->addChild(New.Master);
}

// create object and set description.
// be sure to validate it!
ItemDSO::ItemDSO(dpf *app) : dtkAugment("ItemDSO")
{
	//Am I controlling something
	Control = false;

	//Pointer to dpf application
	App = app;

	//Create the communication object
	Data = new Communicator("ITEM_MESSAGE");

	//Get the hostname of the machine
	unsigned int Size = 64;
	char* Hostname = new char[Size];
	gethostname(Hostname, Size);
	HName = string(Hostname);

	//Connect to the item server if my machine is not the server
	string Server = string(getenv("DPF_SHARED_SERVER"));
	if (Server != HName)
		Data->SetRemote(Server);

	//Validate object
	validate();
	string BaseName = string(getenv("DPF_BASEDIR"));
	printf("Base name is %s\n", BaseName.c_str());
	string DPFDATADIR = string(getenv("DPF_DATA_DIR"));
	printf("Data dir is %s\n", DPFDATADIR.c_str());
        FilePath = ".:./models:../models:" + string(BaseName) + "/DataFiles/avatars:"  + string(BaseName) + "/DataFiles/radar:" + DPFDATADIR;
	printf("File path is %s\n", FilePath.c_str());
}


// destructor
ItemDSO::~ItemDSO(void) 
{
 	delete Data;
}
void ItemDSO::AssociateItems(Message* AMessage)
{
	//Position for the master object
	int MasterPos = -1;
	int ChildPos  = -1;
	string Parent = AMessage->GetString();
	string Child  = AMessage->GetString();
	//Find the master objects position
	for (unsigned int i=0; i< Items.size(); i++)
	{
		if (Items[i].Name == Parent)
			MasterPos = i;
	}
	for (unsigned int i=0; i< Items.size(); i++)
	{
		if (Items[i].Name == Child)
		{
			ChildPos = i;
		}
	}
	if (MasterPos == -1 || ChildPos == -1)
	{
		printf("Association did not work\n");
		return;
	}
	//Take the node out of the world
	App->display()->world()->removeChild(Items[ChildPos].Master);
	//Add the child information to the parent
	Items[MasterPos].ChildList.push_back(Items[ChildPos]);
	//Add the child to the parent dcs
	Items[MasterPos].Children->addChild(Items[ChildPos].Master);
	//Erase the child
	Items.erase(Items.begin() + ChildPos);
}
void ItemDSO::DisassociateItems(Message* AMessage)
{
	//The master node
	string Master = AMessage->GetString();
	//The child node to free
	string Child  = AMessage->GetString();
	//Find parent, then child, then put child back at world level
	for (unsigned int i=0; i< Items.size(); i++)
	{
		//Found the parent
		if (Items[i].Name == Master)
		{
			//Go through children
			for(unsigned int j=0; j< Items[i].ChildList.size(); j++)
			{
				//Found the child
				if (Items[i].ChildList[j].Name == Child)
				{
					//Take the child out of the parent dcs
					Items[i].Children->removeChild(Items[i].ChildList[j].Master);
					//Add the child back to the world
					App->display()->world()->addChild(Items[i].ChildList[j].Master);
					//Add back to parent list
					Items.push_back(Items[i].ChildList[j]);	
					//Take out of the child list of the parent
					Items[i].ChildList.erase(Items[i].ChildList.begin() + j);
				}
			}
		}
	}
}

void ItemDSO::AddItem(Message* AMessage)
{
	//Temporary item
	Item Temp; 
	//Get the name of the object to load into the world
	Temp.Name = AMessage->GetString();
	printf("Making item %s\n", Temp.Name.c_str());
	//Setup the node
	ConstructItem(Temp);
	printf("Constructed node ok\n");
	//Actually load the object into the world
	LoadItem(Temp);
	printf("Loaded item ok\n");
}

void ItemDSO::LoadItem(Item& AnItem)
{
	//Set the correct path for the item to load 
	pfFilePath(FilePath.c_str());
	//Tell the gui what to add to the list of files loaded
	pfNode* Temp = pfdLoadFile(AnItem.Name.c_str());
	AnItem.Parent->addChild(Temp);
	if (Temp == NULL)
	{
		//Put in grabber code here
		printf("Object not found, please make sure you 
		        have the object on your computer in the appropriate
			area before you run this application\n");
		return;
	}
	//Add the node to the world
	App->display()->world()->addChild(AnItem.Master);
	//Add the item to the item list
	Items.push_back(AnItem);
}

void ItemDSO::Grant(Message* Temp)
{
	string Object = Temp->GetString();
	for (unsigned int i=0; i< Items.size(); i++)
	{
		if (Items[i].Name == Object)
		{
			Control = true;
			ControlledItem.Name   = Items[i].Name;
			ControlledItem.Master =  Items[i].Master;
		}
	}
}
void ItemDSO::Revoke(Message* Temp)
{
	if (Temp->GetString() == ControlledItem.Name)
	{
		Control = false;
	}
}
bool ItemDSO::ItemMoved()
{
	return false;	
}

void ItemDSO::WritePosition()
{
	
}

void ItemDSO::LoadItems(Message* Temp)
{
	int Num1 = Temp->GetInt();
	for (int i=0; i < Num1; i++)
	{
		AddItem(Temp);
		//Add in associate code here
		int Num = Temp->GetInt();
		for (int j=0; j< Num; j++)
		{
			AddItem(Temp);
		}
	}
}

void ItemDSO::LoadURL(Message* Temp)
{
	AddItem(Temp);
}

void ItemDSO::MoveObject(Message* Temp)
{
	string Name = Temp->GetString();
	for (unsigned int i=0; i<Items.size(); i++)
	{
		if (Items[i].Name == Name)
		{
			Items[i].Master->setTrans(Temp->GetFloat(), Temp->GetFloat(), Temp->GetFloat());

		}
	}
}

void ItemDSO::RemoveItem(const string& Name)
{
	printf("Supposed to remove item %s\n", Name.c_str());
}

//Every hundred frames output the user list to the collaborative gui
int ItemDSO::postFrame(void) 
{
	//Temporary message
	Message* Temp;
	//Get temporary message
	Temp = Data->GetMessage();
	//No messages to process
	if (Temp == NULL)
		return DTK_CONTINUE;
	//Message is not for me
	if(Temp->GetTo() != HName )
		return DTK_CONTINUE;
	//Add an item to the world
	if (Temp->GetMessage() == "ADD_ITEM")
	{
		AddItem(Temp);
	}
	//Remove an item from the world
	if (Temp->GetMessage() == "REMOVE_ITEM")
	{
		RemoveItem(Temp->GetString());
	}
	//I have control of an object!!!
	if (Temp->GetMessage() == "CONTROL_GRANTED")
	{
		Grant(Temp);
	}
	//Release control of an object that I control
	if (Temp->GetMessage() == "CONTROL_RELEASED")
	{
		Revoke(Temp);
	}
	//Load a list of items to the world (used on startup)
	if (Temp->GetMessage() == "LOAD_ITEMS")
	{
		LoadItems(Temp);
	}
	//URL for an object to be loaded
	if (Temp->GetFrom() == HName && Temp->GetMessage() == "ITEM_DOWNLOADED")
	{
		LoadURL(Temp);
	}
	//Load something into the world
	if (Temp->GetMessage() == "POSITION_IS")
	{
		MoveObject(Temp);
	}
	if (Temp->GetMessage() == "ASSOCIATE_ITEMS")
	{
		AssociateItems(Temp);
	}
	if (Temp->GetMessage() == "DISASSOCIATE_ITEMS")
	{
		DisassociateItems(Temp);
	}
	if (Control && ItemMoved())
		WritePosition();
	return DTK_CONTINUE;
}


/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new ItemDSO(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
