//
// DIVERSE graphics interface -- Performer ---- version 2.0.5
// released: Sat May 18 10:54:26 EDT 2002
// The DIVERSE HOMPAGE is at: www.diverse.vt.edu
// 
// Copyright (C) 2000, 2001, 2002  John Kelso, Lance Arsenault and Ron Kriz
// 
// This software, DIVERSE graphics interface -- Performer (DPF),
// is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License (LGPL) as
// published by the Free Software Foundation; either version 2.1 of
// the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this software, in the top level source directory in
// a file named "COPYING"; if not, see it at:
// http://www.gnu.org/copyleft/lesser.html or write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
//
//  This is an example program for the dpfIntersection class.  It places
//  a gnomon at the location of the object that the wand is pointing at.
//  It does this by projecting a ray from the position and orientation of the
//  wand, and finding the intersection point between this ray and the
//  scene geometry.
#include <stdlib.h>
#include <math.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>

class wandIntersection : public dtkAugment
{
public:

  wandIntersection(dpf *app) ;
  int postConfig(void) ;   
  int preFrame(void) ;   

private:

  dpf *app ;
  dpfDisplay *display ;
  dtkInLocator *wand;
  float wandPos[6] ;
  dpfIntersection* intersect;
  float gnomonPos[6];

  //  Shared memory segment for the local intersection point
  //  3 floats x,y,z
  dtkSharedMem* shm;
} ; 


wandIntersection::wandIntersection(dpf *app_in) :
  dtkAugment("wandIntersection")
{
  app = app_in ;
  setDescription("adds a gnomon object at object "
		 "that the wand is pointing at") ;
  // dtkAugment starts out invalid.
  int c;
  for (c=0; c<6; c++) {
    wandPos[c] = 0.0f;
    gnomonPos[c] = 0.0f;
  }
  //  Get the gnomon out of the way initially
  gnomonPos[1] = -100.0;

  shm = new dtkSharedMem (3 * sizeof(float), "localPoint");
  if (!shm || shm->isInvalid() || shm->getSize() < (3*sizeof(float))) {
    printf ("could not get shared mem\n");
  }

  validate() ; 
}


int wandIntersection::postConfig(void)
{
  if (!(display = app->display())) 
    return DTK_CALLBACK_ERROR ;
  if (!display->scene()) 
    return DTK_CALLBACK_ERROR ;

  //  Initialize the intersection detection.  Segment 0 will contain a line
  //  segment starting at the wand position and extending 100 units in the 
  //  direction of the wand.  This will be in scene coordinates.
  intersect = new dpfIntersection();
  intersect->enableSegment(0);
  intersect->setRoot(display->scene());
  intersect->setSegmentLength(0, 100);
  
  //  Load the gnomon model
  char *path = (char *) pfGetFilePath() ;
  if(path)
    // copy path from static memory to the heap.
    path = dtk_strdup(path);
  pfFilePath(".:"DPF_DATA_DIR);
  dpfDCS *dcs =
    new dpfDCS(gnomonPos, /* position */
		 0.03f, /* overall scale */
		 /* file       scale xyz     position xyz  rotation hpr */
		 "redSphere.pfb", DPF_IDENTITY,
		 NULL);

  if(path)
    {
      pfFilePath(path);
      free(path);
    }
  //  Disable intersection calculation for the gnomon model
  intersect->noIntersection(dcs);
  display->scene()->addChild(dcs) ;
  wand = (dtkInLocator *) app->get("wand", DTKINLOCATOR_TYPE);
  if(!wand) return DTK_CALLBACK_ERROR ;

  return DTK_REMOVE_CALLBACK ;
}


int wandIntersection::preFrame(void)
{
  wand->read(wandPos);

  //  Take the hpr from the wand and make a vector out of it.  Do this by
  //  multiplying the vector {0, 1, 0} by a matrix formed with the hpr 
  //  Euler rotation
  pfMatrix mat;
  mat.makeEuler(wandPos[3], wandPos[4], wandPos[5]);
  pfVec3 vec(0, 1, 0);
  vec.xformVec(vec, mat);

  //  Set up segment 0 and do the intersection
  intersect->setSegmentPos(0, wandPos[0], wandPos[1], wandPos[2]);
  intersect->setSegmentDir(0, vec);
  intersect->doIntersection();

  //  If we got a hit, set the gnomon position to be the position of the
  //  intersection point
  if (intersect->segmentHit(0)) {
    pfVec3 intPoint;
    intersect->getXFormPoint(0, &intPoint);
    gnomonPos[0] = intPoint[0];
    gnomonPos[1] = intPoint[1];
    gnomonPos[2] = intPoint[2];

    //  Put the local intersection point in shared mem
    intersect->getLocalPoint(0, &intPoint);
    float point[3];
    point[0] = intPoint[0];
    point[1] = intPoint[1];
    point[2] = intPoint[2];
    shm->write(point, 3*sizeof(float));
  }
  else {
    gnomonPos[1] = -100;
    shm->write(gnomonPos, 3*sizeof(float));
  }
  
  return DTK_CONTINUE ;
}


/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new wandIntersection(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}


