#include <stdio.h>
#include <dpf.h>
#include <Performer/prmath.h>
#include <dtk.h>
#include <dtk/dtkDSO_loader.h>
#include <string>
#include <vector>
using namespace std;
#include "Communicator.h"
// example generic dtkAugment DSO modified to be a file loader for 
// any set world
// Notice how simple it is to take an example program and modify it

//Used in moving the object
#define DEAD_ZONE (0.05f)
#define DEFAULT_SPEED 16.0f

//An items representation in the world
struct Item
{
	string Name;                     //The description of the item
	pfDCS* Master;                   //Node that is added to the world
	pfDCS* Parent;                   //Node that holds the item
	pfDCS* Children;                 //Node that handles all the child pointers
	vector<Item*> ChildList;          //List of children
	float LastKnownPosition[6];
};

//Represents information on an object that we need to watch
struct WatchItem
{
	dtkSharedMem* Segment;
	string        Computer;
	string        Item;
};

class ItemDSO : public dtkAugment 
{
public:

  ItemDSO(dpf *app) ;
  ~ItemDSO(void) ;
  //During the post frame step handle everything
  int postFrame(void);
  int postConfig(void);

private:

  //Update the position of the object
  void UpdatePosition(Message* Temp);
  //Handle the messages for the gui
  int HandleMessages();
  //Watch an item
  void WatchAnItem(const string& ComputerName, const string& Object);
  //UnWatch an item
  void UnWatchItem(const string& ComputerName, const string& Object);
  //For moving the objects
  void MoveObjects();
  //For releasing control of an object
  void ReleaseControl(Message* Temp);
  //For handling control of objects
  void ControlMessage(Message* Temp);
  //For handling mouse input
  void HandleMouse();
  //Clean everything out
  void Cleanup();
  //Add an item into the world
  void AddItem(Message* AMessage);
  //Remove an item to the world
  void RemoveItem(const string& Name);
  //Handles the actual loading of an item in the world
  void LoadItem(Item* AnItem, const string& URL);
  //Handles the actual un-loading of an item in the world
  void UnLoadItem(Message* AMessage);
  //Load items into the world (at startup)
  void LoadItems(Message* AMessage);
  //Load a specific item located at a url into the world
  void LoadURL(Message* AMessage);
  //Has the item that I have control over moved?
  bool ItemMoved();
  //Write the position of my controlled item to the world
  void WritePosition();
  //Associate items
  void AssociateItems(Message* AMessage);
  //Disassociate items
  void DisassociateItems(Message* AMessage);
  //Setup all the pointers for the item
  void ConstructItem(Item* New);
  //Set the mode of an objects mainpulation
  //By default use the mouse buttons
  //You can set the first button to do only rotate / pan / zoom also
  void SetMode(Message* AMessage);
  //Are we setting a rotation mode on or off
  void SetRotMode(Message* AMessage);
  //Are any modes set?
  bool Modes(); 
  //Set the hpr / xyz increment
  void SetIncrement(Message* AMessage);
  //Set the navigation on and off during item manipulation
  void NavMessage(Message* AMessage);
  //Set the position of an object
  void SetPosition(Message* AMessage);
  //Jump to an items position
  void Jump(Message* AMessage);
  //Highlight an item
  void Highlight(Message* AMessage);
  //Remove an item from the watched list
  void RemoveWatched(const string& Name);
  //Return to normal manipulation
  void NormalManip();

  //Quick pointer to the item I control
  Item* ControlledItem;

  //For mouse input
  dtkInValuator* Pointer;


  //The application that we use
  dpf* App;
  //The communications object
  Communicator* Data;
  //The hostname of the machine
  string HName;
  //The list of items
  vector<Item*> Items;
  vector<WatchItem*> Watched;
  //The file path for loading objects
  string FilePath;
  //Am I controlling something
  bool Control;
  //Used for throttling move messages
  dtkTime* Throttle;
  //Time through one loop iteration
  long double ElaspedTime;
  //Are we supposed to be watching items?
  bool Watch;
  //My segment of shared memory that I write positions to 
  dtkSharedMem* MySegment;
  //The class used to connect to remote machines
  dtkClient* RemoteConnector;
  //The mouse button for the application
  dtkInButton*     Button;
  //The record used to get events out of 
  dtkRecord*       Record;
  //Temp event
  dtkRecord_event* Event;
  //Used to keep track of which mouse button is being held down
  int Pressed ;
  //Used to keep track of what the left button does if a specific action has been chosen
  bool Rotate;
  bool RotX;
  bool RotY;
  bool RotZ;
  bool Pan;
  bool Zoom;
  //How much are we supposed to increment the items position by
  float XYZIncrement;
  float HPRIncrement;
  //Are we going to navigate during item manipulation?
  bool Navigate;
  //Used for highlighting an object
  pfDCS* Highlighter;
  pfNode* Ghost;
  //Is a highlight in place?
  bool TurnMeOff;
  //Number of frames to keep the highlight on
  int Count;
};


bool ItemDSO::Modes()
{
	return (Rotate || Pan || Zoom);
}
void ItemDSO::ConstructItem(Item* New)
{
	//Create nodes
	New->Master   = new pfDCS;
	New->Parent   = new pfDCS;
	New->Children = new pfDCS;
	//Put nodes in proper relationship
	New->Master->addChild(New->Parent);
	New->Master->addChild(New->Children);
	//Display items in the world
	App->display()->world()->addChild(New->Master);
}

// create object and set description.
// be sure to validate it!
ItemDSO::ItemDSO(dpf *app) : dtkAugment("ItemDSO")
{
	TurnMeOff = false;
	Count =0;
	//Basically initialize the mess of variables, connect to the outside world, etc!
	Navigate = false;
	XYZIncrement = 1;
	HPRIncrement = 1;
	Rotate = false;
	RotX = false;
	RotY = false;
	RotZ = false;
	Pan = false;
	Zoom = false;
	Pressed = -1;
	Button = static_cast<dtkInButton *>(app->get("buttons", DTKINBUTTON_TYPE));
	if (Button == NULL)
		printf("Could not get the mouse button connected to shared memory, do not trust program results\n");
	else
	{
		//Button->queue();
		Record = app->record();
		Button->queue();
	}
	RemoteConnector = new dtkClient;
	Watch = false;
	ElaspedTime = 0;
	Throttle = new dtkTime();

	ControlledItem = NULL;
	//Am I controlling something
	Control = false;

	//Pointer to dpf application
	App = app;

	//Create the communication object
	Data = new Communicator("ITEM_MESSAGE");

	//Get the hostname of the machine
	unsigned int Size = 64;
	char* Hostname = new char[Size];
	gethostname(Hostname, Size);
	HName = string(Hostname);

	string TempName = HName + "-ITEM_POSITION";
	MySegment = new dtkSharedMem(sizeof(float)*6, TempName.c_str());
	if (!MySegment || MySegment->isInvalid() || MySegment->getSize() < 6*sizeof(float))
		printf("Could not create shared memory segment, do not trust results\n");

	//Connect to the item server (messages) if my machine is not the server
	string Server = string(getenv("DPF_SHARED_SERVER"));
	if (Server != HName)
		Data->SetRemote(Server);

	//Validate object
	validate();
	
	//Load env configs
	char* Test = getenv("DPF_BASEDIR");
	string BaseName = "";
	if (Test != NULL)
		BaseName = string(Test);
	else
		BaseName = "";
	printf("Item Base name is %s\n", BaseName.c_str());
	string Data1 = "";
	char* Temp = getenv("DPF_DATA_DIR");
	if (Temp == NULL)
		Data1 = "";
	else 
		Data1 = string(Temp);
	printf("Data dir is %s\n", Data1.c_str());
        FilePath = ".:./models:../models:" + string(BaseName) + "/DataFiles/avatars:"  + string(BaseName) + "/DataFiles/radar:" + Data1 + string(DPF_DATA_DIR);
	printf("File path is %s\n", FilePath.c_str());
	pfFilePath(FilePath.c_str());

	//Make the mouse pointer
	Pointer = static_cast<dtkInValuator *> (App->get("pointer", DTKINVALUATOR_TYPE));
	  if(!Pointer || Pointer->number() != 2) 
		  printf("Pointer did not work\n");
	Message Load;
	//Get a list of items that I am supposed to load in the world
	Load.SetTo("ITEM_SERVER");
	Load.SetFrom(HName);
	Load.SetMessage("DSO_LIST_ITEMS");
	Data->SendMessage(Load);
	Ghost = NULL;
	Highlighter = NULL;
}

int ItemDSO::postConfig(void)
{
	//Set up the highlighted object
	return DTK_CONTINUE;
}

//Get rid of everything in the scene graph
void ItemDSO::Cleanup()
{
	delete Throttle;
	delete MySegment;
 	delete Data;
	for (unsigned int i=0; i<Items.size(); i++)
	{
		App->display()->world()->removeChild(Items[i]->Master);
		for (unsigned int j=0; j<Items[i]->ChildList.size(); j++)
		{
			delete Items[i]->ChildList[j];
		}
		delete Items[i];
	}
}


// destructor
ItemDSO::~ItemDSO(void) 
{
	Cleanup();
}
//Associate two items together
void ItemDSO::AssociateItems(Message* AMessage)
{
	//Position for the master object
	int MasterPos = -1;
	int ChildPos  = -1;
	string Parent = AMessage->GetString();
	string Child  = AMessage->GetString();
	printf("Associating parent %s with child %s\n", Parent.c_str(), Child.c_str());
	//Find the master objects position
	for (unsigned int i=0; i< Items.size(); i++)
	{
		if (Items[i]->Name == Parent)
			MasterPos = i;
	}
	for (unsigned int i=0; i< Items.size(); i++)
	{
		if (Items[i]->Name == Child)
		{
			ChildPos = i;
		}
	}
	//Error conditions
	if (MasterPos == -1 || ChildPos == -1 || MasterPos == ChildPos)
	{
		printf("Association did not work\n");
		return;
	}
	//Get the matrix from the parent
	const pfMatrix* MasterMat = Items[MasterPos]->Master->getMatPtr();	
	//Make acopy of it that isn't const
	pfMatrix TempMat;
	TempMat = *(MasterMat);
	//Finally make the inverse of the matrix
	pfMatrix FullMat;
	FullMat.invertFull(TempMat);
	//Now multiply the childs matrix by the inverse of the master matrix so that
	//when we associate them everything is happy in the world
	const pfMatrix* ChildMat = Items[ChildPos]->Master->getMatPtr();
	pfMatrix Paranoia;
	Paranoia = *(ChildMat) * FullMat;
	//TempMat = TempMat * FullMat;
	//Take the node out of the world
	App->display()->world()->removeChild(Items[ChildPos]->Master);
	//Add the child to the parent dcs
	Items[MasterPos]->Children->addChild(Items[ChildPos]->Master);
	//Add the child information to the parent
	Items[MasterPos]->ChildList.push_back(Items[ChildPos]);
	//Set the inverse matrix
	Items[MasterPos]->ChildList[Items[MasterPos]->ChildList.size() -1]->Master->setMat(Paranoia);
	//Erase the child from the parent list
	Items.erase(Items.begin() + ChildPos);
	printf("Finished association\n");
}
//Dissassociate items from one another
void ItemDSO::DisassociateItems(Message* AMessage)
{
	//The master node
	string Child = AMessage->GetString();
	//Find parent, then child, then put child back at world level
	printf("DisAssociating Child %s \n", Child.c_str());
	for (unsigned int i=0; i< Items.size(); i++)
	{
			//Go through children
			for(unsigned int j=0; j< Items[i]->ChildList.size(); j++)
			{
				//Found the child
				if (Items[i]->ChildList[j]->Name == Child)
				{
					printf("Found parent and child, disassociating\n");
					//Get the translation of the child
					//before it leaves the parent
					//so it doesn't go flying off into lala land
					const pfMatrix* Bleh = Items[i]->Master->getMatPtr();
					pfMatrix Temp = *Bleh;
					//Take the child out of the parent dcs
					Items[i]->Children->removeChild(Items[i]->ChildList[j]->Master);
					//Add the child back to the world
					App->display()->world()->addChild(Items[i]->ChildList[j]->Master);
					const pfMatrix* ChildBleh = Items[i]->ChildList[j]->Master->getMatPtr();
					Temp = *(ChildBleh) * Temp;
					//Set the matrix back to what it was
					Items[i]->ChildList[j]->Master->setMat(Temp);

					//Add back to parent list
					Items.push_back(Items[i]->ChildList[j]);	
					//Take out of the child list of the parent
					Items[i]->ChildList.erase(Items[i]->ChildList.begin() + j);
				}
			}
	}
}

//Add an item to the world (calls 2 other functions that do specific tasks)
void ItemDSO::AddItem(Message* AMessage)
{
	//Temporary item
	Item* Temp = new Item; 
	//Get the name of the object to load into the world
	Temp->Name = AMessage->GetString();
	string Bleh = AMessage->GetString();
	printf("Making item %s\n", Temp->Name.c_str());
	//Setup the node
	ConstructItem(Temp);
	printf("Constructed node ok\n");
	//Actually load the object into the world
	LoadItem(Temp, Bleh);
	printf("Loaded item ok\n");
}

//Do the loading of an item with this code
void ItemDSO::LoadItem(Item* AnItem, const string& URL)
{
	//Set the correct path for the item to load 
	pfFilePath(FilePath.c_str());
	printf("Set file path\n");
	//Tell the gui what to add to the list of files loaded
	pfNode* Temp = pfdLoadFile(AnItem->Name.c_str());
	printf("Loaded item\n");
	printf("Added to parent\n");
	if (Temp == NULL)
	{
		if (URL != "DOWNLOADED")
		{
			printf("Could not find item, attempting to download it\n");
			Message Test;
			Test.SetTo("ITEM_SERVER");
			Test.SetFrom(HName);
			Test.SetMessage("NEEDFILE");
			Test.AddString(AnItem->Name.c_str());
			Data->SendMessage(Test);
			App->display()->world()->removeChild(AnItem->Master);
			/*
			delete AnItem->Parent;
			delete AnItem->Children;
			delete AnItem->Master;
			*/
			return;
		}
		else
		{
			printf("Could not load file %s after trying to download it\n", AnItem->Name.c_str());
			return;
		}
	}
	printf("Apparently I have a good file :)\n");
	AnItem->Parent->addChild(Temp);
	//Add the item to the item list
	Items.push_back(AnItem);
	printf("Pushed item to list\n");
}

//Control message sent
void ItemDSO::ControlMessage(Message* Temp)
{
	printf("Doing a control message\n");
	//Was control granted or denied?


	string Result = Temp->GetString();
	printf("Result is %s\n", Result.c_str());
	//If it was denied then do nothing
	if (Result == "DENIED")
	{
		printf("Denied so dying\n");
		return;
	}
	//Get the item I was granted control of 
	string Item   = Temp->GetString();
	printf("Working on item %s, now searching list of size %d\n", Item.c_str(), Item.size());
	//Go through each item and see if it is the one we want
	for (unsigned int i=0; i<Items.size(); i++)
	{
		//Found it 
		if (Items[i]->Name == Item)
		{
			printf("Found item %s\n", Item.c_str());
			//Setup variables
			ControlledItem = Items[i];
			Control = true;
			MySegment->write(Items[i]->LastKnownPosition);
			Throttle->delta();
			dtkNavList* NavList = (dtkNavList*)App->check("dtkNavList", DTKNAVLIST_TYPE);
			NavList->current()->deactivate();
			return;
		}
		//Go through the children and see if they are the item we want
		for (unsigned int j=0; j< Items[i]->ChildList.size(); j++)
		{
			//Found it
			if (Items[i]->ChildList[j]->Name == Item);
			{
				//Setup variables
				ControlledItem = Items[i]->ChildList[j];
				MySegment->write(Items[i]->ChildList[j]->LastKnownPosition);
				dtkNavList* NavList = (dtkNavList*)App->check("dtkNavList", DTKNAVLIST_TYPE);
				NavList->current()->deactivate();
				Control = true;
				Throttle->delta();
				return;
			}
		}
	}
}

//Release control message handler
void ItemDSO::ReleaseControl(Message* Temp)
{
	//Get name of object released
	string Name = Temp->GetString();
	//If we have an item that is controlled
	if (ControlledItem != NULL)
	{
		//Is our controlled item the right item?
		//If so, release control of it
		if (ControlledItem->Name == Name)
		{
			Message Test;
			Test.SetTo("ITEM_SERVER");
			Test.SetFrom(HName);
			Test.SetMessage("UPDATE_KNOWN_POSITION");
			Test.AddString(Name);
			float Holder[6];
			MySegment->read(Holder);
			for (int i=0; i<6; i++)
			{
				//Save the last known position of an item
				ControlledItem->LastKnownPosition[i] = Holder[i];
				Test.AddFloat(Holder[i]);
			}
			Data->SendMessage(Test);
			Control = false;
			ControlledItem = NULL;
			RotX = false;
			RotY = false;
			RotZ = false;
			Zoom = false;
			Pan = false;
			Rotate = false;
			Navigate = false;
			dtkNavList* NavList = (dtkNavList*)App->check("dtkNavList", DTKNAVLIST_TYPE);
			NavList->current()->activate();
		}
	}
}
bool ItemDSO::ItemMoved()
{
	return false;	
}

void ItemDSO::WritePosition()
{
	
}

void ItemDSO::LoadItems(Message* Temp)
{
	int Num1 = Temp->GetInt();
	printf("Loading %d items\n", Num1);
	for (int i=0; i < Num1; i++)
	{
		printf("Adding main parent\n");
		AddItem(Temp);
		bool WatchMe = (bool) Temp->GetInt();
		if (WatchMe)
		{
			WatchAnItem(Temp->GetString(), Items[i]->Name);
			printf("Watching this parent item\n");
		}
		int Num = Temp->GetInt();
		printf("The item has %d children\n", Num);
		for (int j=0; j< Num; j++)
		{
			AddItem(Temp);
			int ChildPos  = Items.size() - 1;
			int MasterPos = Items.size() - 2;
			Message Faker;
			Faker.AddString(Items[MasterPos]->Name);
			Faker.AddString(Items[ChildPos]->Name);
			bool WatchMe = (bool) Temp->GetInt();
			if (WatchMe)
			{
				printf("Watching this child item\n");
				WatchAnItem(Temp->GetString(), Items[Items.size()-1]->Name);
			}
			AssociateItems(&Faker);
			//Add in associate code here
		}
	}
}

void ItemDSO::LoadURL(Message* Temp)
{
	AddItem(Temp);
}

void ItemDSO::RemoveItem(const string& Name)
{
	printf("Supposed to remove item %s\n", Name.c_str());
	for (unsigned int i=0; i< Items.size(); i++)
	{
		//Found it
		if (Items[i]->Name == Name)
		{
			if (Control && ControlledItem->Name == Name)
			{
				Control = false;
				Zoom = false;
				Pan = false;
				Rotate = false;
				RotX = false;
				RotY = false;
				RotZ = false;
				Navigate = false;
				dtkNavList* NavList = (dtkNavList*)App->check("dtkNavList", DTKNAVLIST_TYPE);
				NavList->current()->activate();
				RemoveWatched(Name);
			}
			printf("Found item %s\n", Name.c_str());
			App->display()->world()->removeChild(Items[i]->Master);
			printf("Removed item from world\n");

			for (unsigned int j=0; j< Items[i]->ChildList.size(); j++)
			{
				printf("Freeing child %d\n", j);
				//Remove child from parents
				Items[i]->Master->removeChild(Items[i]->ChildList[j]->Master);
				//Add child back to the main listing 
				Items.push_back(Items[i]->ChildList[j]);
				//Add child to the world
				App->display()->world()->addChild(Items[i]->ChildList[j]->Master);
			}
			//Cleanup pointers
			printf("Ok 1\n");
			delete Items[i]->Parent;
			printf("Ok 2\n");
			delete Items[i]->Children;
			printf("Ok 3\n");
			delete Items[i]->Master;
			printf("Ok 4\n");
			Items.erase(Items.begin() + i);
			printf("Ok 5\n");
		}
		//Check children
		else
		{
			for (unsigned int j=0; j< Items[i]->ChildList.size(); j++)
			{
				if (Items[i]->ChildList[j]->Name == Name)
				{ //Remove child from parents
					Control = false;
					Zoom = false;
					Pan = false;
					Rotate = false;
					RotX = false;
					RotY = false;
					RotZ = false;
					Navigate = false;
					Items[i]->Master->removeChild(Items[i]->ChildList[j]->Master);
					//Free pointers
					delete Items[i]->ChildList[j]->Parent;
					delete Items[i]->ChildList[j]->Children;
					delete Items[i]->ChildList[j]->Master;

					//Kill child
					Items[i]->ChildList.erase(Items[i]->ChildList.begin() + j);
					dtkNavList* NavList = (dtkNavList*)App->check("dtkNavList", DTKNAVLIST_TYPE);
					NavList->current()->activate();
					RemoveWatched(Name);
				}
			}
		}
	}
}

//Handle mouse movement
void ItemDSO::HandleMouse()
{
	if (Navigate)
		return;
	Throttle->delta();
	float xy[2];
	Pointer->read(xy); // poll the pointer
	//Holds the position of the item
	float Data[6];

	//Initalize
	for (int i=0; i<6; i++)
		Data[i] =0;
	//Figure out where my object is
	MySegment->read(Data);	

	//For the users position
	float UserPos[6];
	//Get the navigation out of dpf
	dtkNavList* NavList = (dtkNavList*)App->check("dtkNavList", DTKNAVLIST_TYPE);
	//Set the users position
	for (int i=0; i<6; i++)
		UserPos[i] = NavList->current()->location.d[i];

	//Temporary mouse button variables
	int Left = -1;
	int Middle = -1;
	int Right = -1;

	if (Control)
	{
		//Get the next event out of the dtk server
		Event = Record->getNextEvent(Button);
		//Figure out what buttons were pressed
		Left   = Button->read(0, Event);
		Middle = Button->read(1, Event);
		Right  = Button->read(2, Event);
		//Initialize the clock
		float Delta = (float) Throttle->delta();
		//Increment used in moving the object
		float Increment = (xy[0] > 0.0f) ? ((xy[0]- DEAD_ZONE) / (1.0f - DEAD_ZONE)): ((xy[0] + DEAD_ZONE) / (1.0f - DEAD_ZONE));
		//Cube it for smoothness
		Increment *= Increment * Increment;
		////Got an event
		//printf("Buttons=(%1.1d %1.1d %1.1d)\n", Left, Middle , Right);
		//Left still pressed down
		if (Pressed == 0 && Left)
		{
			//Regular operation or zooming with the first button
			if (!Modes() || Zoom )
			{
				//The pointer is in the dead zone
				if (fabs(xy[1]) < DEAD_ZONE)
					return;

				//Set an empty coord at the origin
				pfCoord Test;
				Test.xyz.set(0,0,0);
				Test.hpr.set(0,0,0);

				//Turn it into dpf coordinates
				App->display()->worldToDpf(&Test);
				//Dead zone of -.1 to .1 enforced
				//Update the y value of the item we want 
				//mess with
				if (fabs(xy[1]) > DEAD_ZONE)
					Test.xyz[PF_Y] += xy[1] *(Delta * Increment + XYZIncrement);
				//Turn the coordinates back into world
				//coords and then update the position of the
				//item
				App->display()->dpfToWorld(&Test);
				Data[0] += Test.xyz[PF_X];
				Data[1] += Test.xyz[PF_Y];
				Data[2] += Test.xyz[PF_Z];
			}
			//We are supposed to pan with the first button
			if (Modes() && Pan)
			{
				//Set an empty coord at the origin
				pfCoord Test;
				Test.xyz.set(0,0,0);
				Test.hpr.set(0,0,0);
				//Turn it into dpf coordinates
				App->display()->worldToDpf(&Test);
				//Dead zone of -.1 to .1 enforced
				//Update the y value of the item we want 
				//mess with
				if (fabs(xy[1]) > DEAD_ZONE)
					Test.xyz[PF_Z] += xy[1] * (Delta * Increment + XYZIncrement);
				if (fabs(xy[0]) > DEAD_ZONE)
					Test.xyz[PF_X] += xy[0] * (Delta * Increment + XYZIncrement);
				//Turn the coordinates back into world
				//coords and then update the position of the
				//item
				App->display()->dpfToWorld(&Test);
				Data[0] += Test.xyz[PF_X];
				Data[1] += Test.xyz[PF_Y];
				Data[2] += Test.xyz[PF_Z];
			}
			// We are supposed to rotate with the first button
			if (Modes() && Rotate)
			{
				//General rotation
				if (!RotX && !RotY && !RotZ)
				{
					//Set an empty coord at the origin
					pfCoord Test;
					Test.xyz.set(0,0,0);
					Test.hpr.set(0,0,0);
					//Turn it into dpf coordinates
					App->display()->worldToDpf(&Test);
					//Dead zone of -.1 to .1 enforced
					//Update the y value of the item we want 
					//mess with
					if (fabs(xy[1]) > DEAD_ZONE)
						Test.hpr[PF_P] += -1*xy[1] * (Delta * Increment + HPRIncrement);
					if (fabs(xy[0]) > DEAD_ZONE)
						Test.hpr[PF_H] += xy[0] * (Delta * Increment + HPRIncrement);
					//Turn the coordinates back into world
					//coords and then update the position of the
					//item
					App->display()->dpfToWorld(&Test);
					Data[3] += Test.hpr[PF_H];
					Data[4] += Test.hpr[PF_P];
					Data[5] += Test.hpr[PF_R];
				}
				//Rotate about the X, Y, or Z axis
				if (RotX)
				{
					if (xy[0] > 0)
						Data[3] = Data[3] + xy[1] * xy[0] * xy[1] + HPRIncrement ;
					else
						Data[3] = Data[3] + xy[1] * xy[0] * xy[1] - HPRIncrement ;
				}
				if (RotY)
				{
					if (xy[0] > 0)
						Data[4] = Data[4] + xy[1] * xy[0] * xy[1] + HPRIncrement ;
					else
						Data[4] = Data[4] + xy[1] * xy[0] * xy[1] - HPRIncrement ;
				}
				if (RotZ)
				{
					if (xy[0] > 0)
						Data[5] = Data[5] + xy[1] * xy[0] * xy[1] + HPRIncrement ;
					else
						Data[5] = Data[5] + xy[1] * xy[0] * xy[1] - HPRIncrement ;
				}
			}
		}
		//Released the middle button
		if (Pressed == 1 && !Middle)
		{
			Pressed = -1;
		}
		//Just pressed the left button
		if(Pressed == -1 && Left)
		{
			Pressed = 0;
			Throttle->delta();
		}
		//Released the left button
		if (Pressed == 0 && !Left)
		{
			Pressed = -1;
		}
		//Middle still preessed
		if (Pressed == 1 && Middle && !Modes())
		{
			if (!Modes())
			{
				//General rotation
				//Set an empty coord at the origin
				pfCoord Test;
				Test.xyz.set(0,0,0);
				Test.hpr.set(0,0,0);
				//Turn it into dpf coordinates
				App->display()->worldToDpf(&Test);
				//Dead zone of -.1 to .1 enforced
				//Update the y value of the item we want 
				//mess with
				if (fabs(xy[1]) > DEAD_ZONE)
					Test.hpr[PF_P] += -1*xy[1] * (Delta * Increment + HPRIncrement);
				if (fabs(xy[0]) > DEAD_ZONE)
					Test.hpr[PF_H] += xy[0] * (Delta * Increment + HPRIncrement);
				//Turn the coordinates back into world
				//coords and then update the position of the
			//item
				App->display()->dpfToWorld(&Test);
				Data[3] += Test.hpr[PF_H];
				Data[4] += Test.hpr[PF_P];
				Data[5] += Test.hpr[PF_R];
			}
		}
		//Released the middle button
		if (Pressed == 1 && !Middle)
		{
			Pressed = -1;
		}
		//Pressed the middle button
		if (Pressed == -1 && Middle)
		{
			Pressed = 1;
			Throttle->delta();
		}
		//Right button pressed
		if (Pressed == -1 && Right)
		{
			Pressed = 2;
		}
		//Right button released
		if (Pressed == 2 && !Right)
		{
			Pressed = -1;
		}
		//Right button still held
		if (Pressed == 2 && Right && !Modes())
		{
			if (Modes())
				return;
			//Set an empty coord at the origin
			pfCoord Test;
			Test.xyz.set(0,0,0);
			Test.hpr.set(0,0,0);
			//Turn it into dpf coordinates
			App->display()->worldToDpf(&Test);
			//Dead zone of -.1 to .1 enforced
			//Update the y value of the item we want 
			//mess with
			if (fabs(xy[1]) > DEAD_ZONE)
				Test.xyz[PF_Z] += xy[1] * (Delta * Increment + XYZIncrement);
			if (fabs(xy[0]) > DEAD_ZONE)
				Test.xyz[PF_X] += xy[0] * (Delta * Increment + XYZIncrement);
			//Turn the coordinates back into world
			//coords and then update the position of the
			//item
			App->display()->dpfToWorld(&Test);
			Data[0] += Test.xyz[PF_X];
			Data[1] += Test.xyz[PF_Y];
			Data[2] += Test.xyz[PF_Z];
		}
		//Finally write the result of the data to shared memory :)
		MySegment->write(Data);
	}
}

//Move an item in the world
void ItemDSO::WatchAnItem(const string& ComputerName, const string& Item)
{
	printf("Watching item %s on computer %s\n", Item.c_str(), ComputerName.c_str());
	//Tell the dso we are watching something
	if (!Watch)
		Watch   = true;
	//Make a new watch item
	WatchItem* NewWatch  = new WatchItem;
	//Set information
	NewWatch->Computer   = ComputerName;
	NewWatch->Item       = Item;
	//Create the shared memory segment to connect to
	string Name     = ComputerName + "-ITEM_POSITION";
	//Make the shared memory segment
	NewWatch->Segment  = new dtkSharedMem(sizeof(float)*6, Name.c_str());
	//Connect to it remotely if it is not local
	if (HName != ComputerName)
		RemoteConnector->connectRemoteSharedMem(Name.c_str(), ComputerName.c_str());
	Watched.push_back(NewWatch);
}

//Make an item not watched
void ItemDSO::UnWatchItem(const string& ComputerName, const string& Item)
{
	//If we only have one object, stop watching it!
	if (Watched.size() == 1)
		Watch = false;

	//Go through each watched object and see if it is the one we want
	for (unsigned int i=0; i< Watched.size(); i++)
	{
		if (Watched[i]->Item == Item)
		{
			//Save the items last known position so we don't get 
			//jumping all over the place if we request control of it :)
			for (unsigned int j=0; j< Items.size(); j++)
			{
				if (Items[i]->Name == Item)
				{
					Watched[i]->Segment->read(Items[i]->LastKnownPosition);
				}
				for (unsigned int k=0; k< Items[i]->ChildList.size(); k++)
				{
					if (Items[i]->ChildList[j]->Name == Item)
					{
						Watched[i]->Segment->read(Items[i]->ChildList[j]->LastKnownPosition);
						
					}
				}
			}
			//Free shared memory segment
			delete Watched[i]->Segment;
			//Erase from watch list
			delete Watched[i];
			//Erase an item from the watched list
			Watched.erase(Watched.begin() + i);
		}
	}
}

//Move the objects in the world to the correct position
void ItemDSO::MoveObjects()
{
	float Data[6];
	string Name = "";
	for (int i=0; i<6; i++)
		Data[i] = 0;
	for (unsigned int a=0; a < Watched.size(); a++)
	{
		Name = Watched[a]->Item;
		Watched[a]->Segment->read(Data);
		//Go through the parents and see if we find the child
		for (unsigned int i=0; i < Items.size(); i++)
		{
			if (Items[i]->Name == Name)
			{
				Items[i]->Master->setTrans(Data[0],Data[1],Data[2]);
				Items[i]->Master->setRot(Data[3],Data[4],Data[5]);
			}
			for (unsigned int j=0; j< Items[i]->ChildList.size(); j++)
			{
				if (Items[i]->ChildList[j]->Name == Name)
				{
					Items[i]->ChildList[j]->Master->setTrans(Data[0],Data[1], Data[2]);
					Items[i]->ChildList[j]->Master->setRot(Data[3],Data[4], Data[5]);
				}
			}
		}
	}
}


//Every hundred frames output the user list to the collaborative gui
int ItemDSO::postFrame(void) 
{
	if (TurnMeOff)
	{
		Count++;
	}
	if (TurnMeOff && Count > 300)
	{
		TurnMeOff = false;
		Count = 0;
		Highlighter->setTravMask(PFTRAV_APP|PFTRAV_CULL|PFTRAV_DRAW,0,
				         PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
		App->display()->world()->removeChild(Highlighter);
	}
	//Handle all of the messages sent to me
	while (HandleMessages() != -1);

	//If we are watching objects then get to moving
	if (Watch)
		MoveObjects();

	//If I have control of an object and it is the right time
	//move it.  Otherwise just handle a bunch of messages :)
	if (Control)
	{
			//Handle mouse input
			HandleMouse();
	}
	//Return
	return DTK_CONTINUE;
}

//Update the position of the item
void ItemDSO::UpdatePosition(Message* Temp)
{
	string Var = Temp->GetString();	
	float Value = Temp->GetFloat();
	float Test[6];
	MySegment->read(Test);

	switch (Var.at(0))
	{
		case 'X':
		{
			Test[0] = Value;
			break;
		}
		case 'Y':
		{
			Test[1] = Value;
			break;
		}
		case 'Z':
		{
			Test[2] = Value;
			break;
		}
		case 'H':
		{
			Test[3] = Value;
			break;
		}
		case 'P':
		{
			Test[4] = Value;
			break;
		}
		case 'R':
		{
			Test[5] = Value;
			break;
		}
	}
	MySegment->write(Test);
}

//Set which mode I am going to mainpulate the object with
//the first mouse button
void ItemDSO::SetMode(Message* AMessage)
{
	string Mode;   
	string MyStatus;
	Mode = AMessage->GetString();
	MyStatus = AMessage->GetString();
	printf("DSO Setting status to be %s %s\n", Mode.c_str(), MyStatus.c_str());
	//Turn navigation off
	if (Navigate)
		Navigate = false;
	//Turn performer navigation off
	dtkNavList* NavList = (dtkNavList*)App->check("dtkNavList", DTKNAVLIST_TYPE);
	NavList->current()->deactivate();

	//Something got turned off so everything gets turned off
	if (MyStatus == "OFF")
	{
		Pan = false;
		Zoom = false;
		Rotate = false;
		RotX = false;
		RotY = false;
		RotZ = false;
		return;
	}
	//Go through and set which variables need to be set
	if (Mode == "PAN")
	{
		Pan = true;
		Zoom = false;
		Rotate = false;
		RotX = false;
		RotY = false;
		RotZ = false;
		return;
	}
	if (Mode == "ROTATE")
	{
		Pan =  false;
		Zoom = false;
		Rotate = true;
		RotX = false;
		RotY = false;
		RotZ = false;
		return;
	}
	if (Mode == "ZOOM")
	{
		Pan =  false;
		Zoom = true;
		Rotate = false;
		RotX = false;
		RotY = false;
		RotZ = false;
		return;
	}
}

//Set which axis we are supposed to rotate around
void ItemDSO::SetRotMode(Message* AMessage)
{
	printf("Trying to rotate something\n");
	if (!Rotate)
		return;
	string Mode = AMessage->GetString();
	string MyStatus = AMessage->GetString();
	printf("Rotating around %s and %s\n", Mode.c_str(), MyStatus.c_str());
	if (MyStatus == "OFF")
	{
		RotX = false;
		RotY = false;
		RotZ = false;
		return;
	}
	if (Mode == "X")
	{
		RotZ = false;
		RotY = false;
		RotX = true;
	}
	if (Mode == "Y")
	{
		RotX = false;
		RotZ = false;
		RotY = true;
	}
	if (Mode == "Z")
	{
		RotX = false;
		RotY = false;
		RotZ = true;
	}
}

void ItemDSO::NavMessage(Message* AMessage)
{
	//Navigation status
	string MyStatus = AMessage->GetString();
	//Get the naviagation
	dtkNavList* NavList = (dtkNavList*)App->check("dtkNavList", DTKNAVLIST_TYPE);
	//Turn it on
	if (MyStatus == "ON")
	{
		NavList->current()->activate();
		Navigate = true;
	}
	//Turn it off
	else
	{
		NavList->current()->deactivate();
		Navigate = false;
	}
}

void ItemDSO::SetPosition(Message* AMessage)
{
	//Get name of object to move
	string Name = AMessage->GetString();
	float Data[6];
	for (int i=0; i<6; i++)
	{
		Data[i] = AMessage->GetFloat();
		printf("%d %f\n", i, Data[i]);
	}
	for (unsigned int i=0; i< Watched.size(); i++)
	{
		if (Watched[i]->Item== Name)
		{
			printf("Item being watched, ignore\n");
			return;
		}
	}
	//Go through the parents and see if we find the child
	for (unsigned int i=0; i < Items.size(); i++)
	{
		if (Items[i]->Name == Name)
		{
			printf("Parent position set\n");
			Items[i]->Master->setTrans(Data[0],Data[1],Data[2]);
			Items[i]->Master->setRot(Data[3],Data[4],Data[5]);
			for (int k=0; k<6; k++)
			{
				Items[i]->LastKnownPosition[k] = Data[k];
			}
		}
		//Go through and see if it is a child that we need to move
		for (unsigned int j=0; j< Items[i]->ChildList.size(); j++)
		{
			if (Items[i]->ChildList[j]->Name == Name)
			{
				for (int k=0; k<6; k++)
				{
					Items[i]->ChildList[j]->LastKnownPosition[k] = Data[k];
				}
				printf("Child position set\n");
				Items[i]->ChildList[j]->Master->setTrans(Data[0],Data[1], Data[2]);
				Items[i]->ChildList[j]->Master->setRot(Data[3],Data[4], Data[5]);
			}
		}
	}
}

//Jump to an items position with a small offset
void ItemDSO::Jump(Message* AMessage)
{
	//Adjusted 5 units back from the object
	pfCoord OffSet;
	OffSet.xyz.set(0,0,0);
	OffSet.hpr.set(0,0,0);
	//Turn it into dpf coordinates

	//Get the navigation list out of the dpf app
	dtkNavList* NavList = (dtkNavList*)App->check("dtkNavList", DTKNAVLIST_TYPE);
	//Get the name of the object we want to jump to
	string Name = AMessage->GetString();
	//Data used to jump to a watched items position
	float Test[6];
	for (unsigned int i=0; i < Watched.size(); i++)
	{
		if (Watched[i]->Item == Name)
		{
			//Read the watched items position
			Watched[i]->Segment->read(Test);
			//Set the coordinate up
			OffSet.xyz.set(Test[0], Test[1], Test[2]);
			OffSet.hpr.set(Test[3], Test[4], Test[5]);
			//Change to dpf coordinates
			App->display()->worldToDpf(&OffSet);
			if (OffSet.xyz[PF_Y] > 0)
				OffSet.xyz[PF_Y] -=100;
			else
				OffSet.xyz[PF_Y] +=100;
			//Change back to world coordinates
			App->display()->dpfToWorld(&OffSet);
			//Turn on navigation so we can jump
			NavList->current()->activate();
			//Go through and goto the specific position
			for (int k=0; k<3; k++)
			{
				NavList->current()->location.d[k] = (OffSet.xyz[k]);
				NavList->current()->location.d[k+3] = (OffSet.hpr[k]);
			}
			NavList->current()->deactivate();
			return;
		}
	}
	//Go through the parent list since it is not a watched item
	for (unsigned int i=0; i< Items.size(); i++)
	{
		if (Items[i]->Name == Name)
		{
			OffSet.xyz.set(Items[i]->LastKnownPosition[0], Items[i]->LastKnownPosition[1], Items[i]->LastKnownPosition[2]);
			OffSet.hpr.set(Items[i]->LastKnownPosition[3], Items[i]->LastKnownPosition[4], Items[i]->LastKnownPosition[5]);
			//Change to dpf coordinates
			App->display()->worldToDpf(&OffSet);
			//Step 10 units back
			if (OffSet.xyz[PF_Y] > 0)
				OffSet.xyz[PF_Y] -=100;
			else
				OffSet.xyz[PF_Y] +=100;

			//Change back to world coordinates
			App->display()->dpfToWorld(&OffSet);
			//Go through and goto the specific position
			for (int k=0; k<3; k++)
			{
				NavList->current()->location.d[k] = (OffSet.xyz[k]);
				NavList->current()->location.d[k+3] = (OffSet.hpr[k]);
			}
			//Go through and goto the specific position
		}
		//Go through the child list of the parent 
		for (unsigned int j=0; j< Items[i]->ChildList.size(); j++)
		{
			if (Items[i]->ChildList[j]->Name == Name)
			{
				//Get Position of item
				OffSet.xyz.set(Items[i]->ChildList[j]->LastKnownPosition[0], Items[i]->ChildList[j]->LastKnownPosition[1], Items[i]->ChildList[j]->LastKnownPosition[2]);
				OffSet.hpr.set(Items[i]->ChildList[j]->LastKnownPosition[3], Items[i]->ChildList[j]->LastKnownPosition[4], Items[i]->ChildList[j]->LastKnownPosition[5]);
				//Change to dpf coordinates
				App->display()->worldToDpf(&OffSet);
				if (OffSet.xyz[PF_Y] > 0)
					OffSet.xyz[PF_Y] -=100;
				else
					OffSet.xyz[PF_Y] +=100;
				//Change back to world coordinates
				App->display()->dpfToWorld(&OffSet);
				//Go through and goto the specific position
				for (int k=0; k<3; k++)
				{
					NavList->current()->location.d[k] = (OffSet.xyz[k]);
					NavList->current()->location.d[k+3] = (OffSet.hpr[k]);
				}
			}
		}
	}
}
void ItemDSO::Highlight(Message* AMessage)
{
	if (TurnMeOff)
		return;
	if (Highlighter == NULL)
		Highlighter = new pfDCS;
	if (Ghost == NULL)
	{
		Ghost = pfdLoadFile("sphere.pfb");
		if (Ghost == NULL)
		{
			printf("Could not find sphere.pfb in the item dso\n");
			return;
		}
		Highlighter->addChild(Ghost);
	}
	//Add the highlight node to the world
	App->display()->world()->addChild(Highlighter);
	//Set the traversal
	Highlighter->setTravMask(PFTRAV_APP|PFTRAV_CULL|PFTRAV_DRAW,1,
	PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
	//Now get the dimensions of the sphere
	pfSphere Sphere;
	string Name = AMessage->GetString();
	for (unsigned int i=0; i< Items.size(); i++)
	{
		if (Items[i]->Name == Name)
		{
			Sphere = dpf_getBSphere(Items[i]->Parent);	
			printf("Highlighting parent %s\n", Name.c_str());
		}
		for (unsigned int j=0; j< Items[i]->ChildList.size(); j++)
		{
			if (Items[i]->ChildList[j]->Name == Name)
			{
				Sphere = dpf_getBSphere(Items[i]->ChildList[j]->Parent);	
				printf("Highlighting child %s\n", Name.c_str());
			}	
		}
	}
	if (Sphere.radius >0.f)
	{
		pfMatrix Mat ;
		Mat.makeTrans(Sphere.center[PF_X], Sphere.center[PF_Y], Sphere.center[PF_Z]) ;
		Mat.preScale(Sphere.radius, Sphere.radius, Sphere.radius, Mat) ;
		Highlighter->setMat(Mat) ;
		Highlighter->setTravMask(PFTRAV_APP|PFTRAV_CULL|PFTRAV_DRAW,1,
		PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
	}
	else
	{
		Highlighter->setTravMask(PFTRAV_APP|PFTRAV_CULL|PFTRAV_DRAW,0,
		PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
	}
	TurnMeOff = true;
	Count =0;
}

//Remove a watched item
void ItemDSO::RemoveWatched(const string& Name)
{
	for (unsigned int i=0; i< Items.size(); i++)
	{
		if (Watched[i]->Item == Name)	
		{
			Watched.erase(Watched.begin() + i);
		}
	}
}
void ItemDSO::NormalManip()
{
	RotX = false;
	RotY = false;
	RotZ = false;
	Pan  = false;
	Rotate = false;
	Zoom = false;
	Navigate = false;
	dtkNavList* NavList = (dtkNavList*)App->check("dtkNavList", DTKNAVLIST_TYPE);
	NavList->current()->deactivate();
}

//Handles messages sent to the dso by the server and mainly by the gui
int ItemDSO::HandleMessages()
{
	//Temporary message
	Message* Temp;
	//Get temporary message
	Temp = Data->GetMessage();
	//No messages to process
	if (Temp == NULL)
		return -1;
	//Message is not for me
	if(Temp->GetTo() != HName )
		return 0;
	//Add an item to the world
	if (Temp->GetMessage() == "ADD_ITEM")
	{
		AddItem(Temp);
	}
	//Remove an item from the world
	if (Temp->GetMessage() == "REMOVE_ITEM")
	{
		RemoveItem(Temp->GetString());
	}
	//I might have control of an object!!!
	if (Temp->GetMessage() == "ITEM_CONTROL")
	{
		ControlMessage(Temp);
	}
	//Load a list of items to the world (used on startup)
	if (Temp->GetMessage() == "DSO_LOAD_ITEMS")
	{
		LoadItems(Temp);
	}
	//URL for an object to be loaded
	if (Temp->GetFrom() == HName && Temp->GetMessage() == "ITEM_DOWNLOADED")
	{
		LoadURL(Temp);
	}
	//Associate two items together
	if (Temp->GetMessage() == "ASSOCIATE")
	{
		AssociateItems(Temp);
	}
	//DisAssociate items from one another
	if (Temp->GetMessage() == "DISASSOCIATE")
	{
		DisassociateItems(Temp);
	}
	//Release control of the items
	if (Temp->GetMessage() == "CONTROL_RELEASED")
	{
		ReleaseControl(Temp);
	}
	//Are we supposed to watch an item?
	if (Temp->GetMessage() == "WATCH_ITEM")
	{
		string Name   = Temp->GetString();
		string Object = Temp->GetString();
		WatchAnItem(Name, Object);
	}
	//Are we supposed to stop watching an item
	if (Temp->GetMessage() == "UNWATCH_ITEM")
	{
		string Name   = Temp->GetString();
		string Object = Temp->GetString();
		UnWatchItem(Name, Object);
	}
	//The local gui has downloaded a file for us
	//Attempt to load it
	if (Temp->GetMessage()== "FILE_DOWNLOADED")
	{
		AddItem(Temp);
	}
	//Update the position of an object
	if (Temp->GetMessage() == "UPDATE_POSITION")
	{
		UpdatePosition(Temp);
	}
	//Set how we are going to manipulate it
	if (Temp->GetMessage() == "SET_MANIPULATION_MODE")
	{
		SetMode(Temp);
	}
	//Set which axis we are going to rotate around (x,y,z)
	if (Temp->GetMessage() == "SET_ROT_MODE")
	{
		SetRotMode(Temp);
	}
	//Turn naviagtaion either on or off
	if (Temp->GetMessage() == "NAVIGATION")
	{
		NavMessage(Temp);
	}
	//Update the xyz increment
	if (Temp->GetMessage() == "XYZ")
	{
		XYZIncrement = Temp->GetFloat();
	}
	//Update the hpr increment
	if (Temp->GetMessage() == "HPR")
	{
		HPRIncrement = Temp->GetFloat();
	}
	//Set the position of the item
	if (Temp->GetMessage() == "ITEM_POSITION")
	{
		SetPosition(Temp);
	}
	//Jump to an items position
	if (Temp->GetMessage() == "JUMP_TO_ITEM")
	{
		Jump(Temp);
	}
	//Highlight an item
	if (Temp->GetMessage() == "HIGHLIGHT_ITEM")
	{
		Highlight(Temp);

	}
	//Set everything back to normal item manipulation
	if (Temp->GetMessage() == "RETURN_TO_NORMAL_MANIPULATION")
	{
		NormalManip();
	}
	delete Temp;

	//
	/*
	if (Temp->GetMessage() == "DIE")
	{
		Cleanup();
	}
	*/
	return DTK_CONTINUE;
}


/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new ItemDSO(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
