///////////// MessageParser.h //////////////////////
//  Used to parse commands, or messages.
////////////////////////////////////////////////////

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//
// All Messages should be in the following form.
// 
// A string of >= 1024 characters
//
// Formatting as follows.
//
// ______ <tab> ____ <tab> ____________ <tab> _______ <tab> ___________________
// (From)       (To)       (Time Stamp)       (see *)       (parameters if any)
//
// * the actual message or command
//
// Examples:
//
// server <tab> pitr.sv.vt.edu <tab> 08:30:00 <tab> disconnect
//
// Class1 <tab> Class2 <tab> 09:30:00 <tab> Set_Speed <tab> 20
// 
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#ifndef MESSAGEPARSER_H
#define MESSAGEPARSER_H

#include <string>
#include <sstream>

using namespace std;

class MessageParser
{
  private:
    string CurrMessage;
    istringstream* istr;
    char buff[1024];
    dtkSharedMem* Receive;

    string from;
    string to;
    string time;
    string message;

  public:
    MessageParser(string ShmName = "");
    ~MessageParser();
    
    int NextMessage();
    
    string getFrom() {return from;}
    string getTo() {return to;}
    string getTime() {return time;}

//  purpose:    To return only the actual message/command of the  entire message.
    string getMessage() {return message;}

    int getInt();
    // Returns everything after the message until the delimiter
    string getLine(char delimiter = '\0');
    string getString();
    char getChar();
    double getDouble();
    
};

MessageParser::MessageParser(string ShmName)
{
  CurrMessage = "";
  istr = NULL;
  
  if(ShmName.length() == 0){
    char* charShmName = getenv("DTK_COMMAND_SHM_NAME");
    if(charShmName == NULL) ShmName = "CommandMessage";
    else ShmName = charShmName;

    if(ShmName.length() == 0) ShmName = "CommandMessage";
  }

  Receive = new dtkSharedMem(1024*sizeof(char),ShmName.c_str());
  if(!Receive || Receive->isInvalid() || Receive->getSize() < (1024*sizeof(char))) {
		printf ("could not get shared mem in MessageParser\n");
  } else {
    printf("M:connected to sharedMem = %s\n", ShmName.c_str());
    Receive->flush();
    Receive->queue();
  }
}

MessageParser::~MessageParser()
{
  delete istr;
  istr = NULL;
}

///////////////////////////////////////////////////////////////////////
//  NextMessage
//
//  Purpose:  If there is a new message on the queue then it will yank
//            it of and start parsing it.  It will pull the structured
//            info out of the message (i.e. the from,to,time,command).
//
//  Post Condition: Warning... if there is a new message on the queue
//            all of the old from,to,time,command will be replaced by
//            the new information. 
//
//  returns:  0 if there is something new on the queue.
//            and -1 if it there isn't.
//            
///////////////////////////////////////////////////////////////////////

int MessageParser::NextMessage()
{
  if(!Receive->qread(buff)) {
//    printf("M: did not read\n");
    return -1;
  }
   
  // Eric 07/18
  if(!buff){
    printf("M:buff NULL\n");
    return -1;
  }
  
  if(istr != NULL){
    delete istr;
    istr = NULL;
  }
  
  istr = new istringstream(buff);
  getline((*istr),from,'\t');
  getline((*istr),to,'\t');
  getline((*istr),time,'\t');
  getline((*istr),message,'\t');

  return 0;
}

int MessageParser::getInt()
{
  int nextInt;
  if(istr)  (*istr)>>nextInt;	  //E
  return nextInt;
}

string MessageParser::getString()
{
  string nextString = "";
  if(istr) (*istr)>>nextString;	  //E
  return nextString;
}

string MessageParser::getLine(char delimiter) 
{
  string nextString = "";
  if(istr) {
//    char ch;
//    istr->get(ch);
    getline( (*istr), nextString, delimiter );
//    delete istr;
//    string temp = "\t" + nextString;
//    istr = new istringstream(temp);
  }

  return nextString;
}

double MessageParser::getDouble()
{
  double nextD;
  if(istr) (*istr)>>nextD;	  //E
  return nextD;
}

char MessageParser::getChar()
{
  char nextChar;
  if(istr) (*istr)>>nextChar;	  //E
  return nextChar;
}

#endif
