#include "fogDso.h"
#include <FL/Fl_Value_Slider.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Color_Chooser.h>
#include <Performer/pr/pfFog.h>

#define DESCRIPTION "Allows users to modify the visibility of the scene," \
		    "by changing the fog values\n"
#define R 0
#define G 1
#define B 2

static float r = .772, g = .807, b = 1.0;
static bool isOn = true;
static Fl_Window* make_window ();

/********************************
 *	  Class Methods
 *******************************/

// constructor 
fogDso::fogDso(dpf *app_in):dtkAugment("fog") {
  setDescription(DESCRIPTION);
  app = app_in;
  fog = NULL;
  validate();
}

// destructor
fogDso::~fogDso() {
  if(guiWin) {
    delete guiWin;
    guiWin = NULL;
  }
}

// initialize everything
int fogDso::postConfig(){
 
  // initialize the fog
  pfEnable(PFEN_FOG);
  
  fog = new pfFog;
  fog->setFogType(PFFOG_VTX_EXP2);
  fog->setColor(0.772, .807, 1.0f);
  fog->setRange(0.0f, 10000.0);
  fog->apply();

  pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_ON);

  // create the gui
  guiWin = make_window();
  if(!guiWin) return DTK_CALLBACK_ERROR;

  return DTK_REMOVE_CALLBACK;
}

// check gui to see of stuff needs 
// to get done each frame
int fogDso::postFrame() {
  Fl::check();
  return DTK_CONTINUE;
}

/**********************************
 *	    Callbacks
 *********************************/

// turns the fog on and off
static void onOff_callback(Fl_Button* onOff, Fl_Value_Slider* s) {
  pfFog *fog = pfGetCurFog();
  //static pfFog *lastFog = NULL;

  if(fog) {

    isOn = !isOn;
//    lastFog = fog;
    pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_OFF);
    
    if(isOn){
      pfEnable(PFEN_FOG);
      onOff->label("Fog On");
  
      fog->setFogType(PFFOG_VTX_EXP2);
      fog->setRange(0.0f,s->value());
      fog->setColor(r,g,b);
      fog->apply();

      pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_ON);
    }else onOff->label("Fog Off");
  }
}

// modifies the opacity of the fog with sliders
static void modifyVisibility_callback(Fl_Value_Slider* s, void*) {
 	
  pfFog* fog = pfGetCurFog();
  
  if(fog && isOn) {
    pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_OFF);

    fog->setRange(0.0f,s->value());
    fog->apply();

    pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_ON);
  }
}

// changes the color of the fog according to rgb values
// from the Fl_Color_Chooser widget
static void modifyColor_callback(Fl_Color_Chooser *color, void*) {
  pfFog* fog = pfGetCurFog();

  if(fog && isOn) {
    pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_OFF);

    r = color->r();
    g = color->g();
    b = color->b();

    fog->setColor(r,g,b);
    fog->apply();

    pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_ON);
  }  
}

// reset the color back to the initial rgb values
static void reset_callback(Fl_Button *, void* color) {
  pfFog* fog = pfGetCurFog();

  if(fog && isOn) {
    pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_OFF);
    ((Fl_Color_Chooser*)color)->rgb((r=.772),(g=.807),(b=1.0));

    fog->setColor(r,g,b);
    fog->apply();
    
    pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_ON);
  }
}

// hides the gui
static void hide_callback(Fl_Widget*, void* w) {
   ((Fl_Window*)w)->iconize();
}

// creates the gui
static Fl_Window* make_window() {
  Fl_Window* w;
  Fl_Color_Chooser *color;
  Fl_Value_Slider* s = NULL;
  { Fl_Window* o = new Fl_Window(300, 230, "Fog Value");
    w = o;
    { Fl_Value_Slider* o = s = new Fl_Value_Slider(5, 25, 290, 25, "Visibility:  ");
      o->type(5);
      o->selection_color(34);
      o->labeltype(FL_ENGRAVED_LABEL);
      o->minimum(50);
      o->step(1);
      o->maximum(50000);
      o->value(10000);
      o->align(FL_ALIGN_TOP_LEFT);
      o->callback((Fl_Callback*)modifyVisibility_callback);
    }
    { Fl_Color_Chooser *o = color = new Fl_Color_Chooser(5,75,290,115, "Fog Colors");
      o->rgb(r,g,b);
      o->labeltype(FL_ENGRAVED_LABEL);
      o->align(FL_ALIGN_TOP_LEFT);
      o->callback((Fl_Callback*)modifyColor_callback);
      o->when(FL_WHEN_CHANGED);     
    } 
    { Fl_Button* o = new Fl_Button(5, 195, 93, 30, "Fog On");
      o->labeltype(FL_ENGRAVED_LABEL);  
      o->callback((Fl_Callback*)onOff_callback, (void*)s);
    }
    { Fl_Button* o = new Fl_Button(103, 195, 93, 30, "Reset");	//5, 195, 100, 30
      o->labeltype(FL_ENGRAVED_LABEL);
      o->callback((Fl_Callback*)reset_callback, (void*)color);
    } 
    { Fl_Button* o = new Fl_Button(201, 195, 93, 30, "Hide");
      o->labeltype(FL_ENGRAVED_LABEL);
      o->callback((Fl_Callback*)hide_callback, (void*) w);
    }
    o->end();
  }

  w->show();

  return w;
}

/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new fogDso(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}

