//
// DIVERSE graphics interface -- Performer ---- version 2.0.5
// released: Sat May 18 10:54:26 EDT 2002
// The DIVERSE HOMPAGE is at: www.diverse.vt.edu
// 
// Copyright (C) 2000, 2001, 2002  John Kelso, Lance Arsenault and Ron Kriz
// 
// This software, DIVERSE graphics interface -- Performer (DPF),
// is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License (LGPL) as
// published by the Free Software Foundation; either version 2.1 of
// the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this software, in the top level source directory in
// a file named "COPYING"; if not, see it at:
// http://www.gnu.org/copyleft/lesser.html or write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
//
//  This is a DSO to demonstrate mouse selection on a desktop system.
//  A gnomon model is placed at the location the mouse is pointing at.
//  This class assumes there is only one dpfScreen and one dpfPWin.
//  It also assumes the field of view does not change for the duration of
//  program.  However, it will still work if the window is resized.

#include <stdlib.h>
#include <math.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>

class mouseSelector : public dtkAugment
{
public:

  mouseSelector(dpf *app) ;
  int postConfig(void) ;   
  int preFrame(void) ;   

private:

  dpf *app ;
  dpfDisplay *display ;
  dpfIntersection* intersect;
  dtkInValuator *pointer;   //  Mouse pointer
  float gnomonPos[6];    // Position of the gnomon model, in scene coordinates
  float h;   // Distance from the viewing window to the user

  //  Shared memory segment for the local intersection point
  //  3 floats x,y,z
  dtkSharedMem* shm;
} ; 


mouseSelector::mouseSelector(dpf *app_in) : dtkAugment("mouseSelector")
{
  app = app_in ;
  setDescription("adds a gnomon at th object that the mouse is pointing at") ;
  // dtkAugment starts out invalid.
  int c;
  for (c=0; c<6; c++) {
    gnomonPos[c] = 0.0f;
  }

  //  Initialize the gnomon to be behind you
  gnomonPos[1] = -100.0f;

  shm = new dtkSharedMem (3 * sizeof(float), "localPoint");
  if (!shm || shm->isInvalid() || shm->getSize() < (3*sizeof(float))) {
    printf ("could not get shared mem\n");
  }

  validate() ; 
}


int mouseSelector::postConfig(void)
{
  if (!(display = app->display())) 
    return DTK_CALLBACK_ERROR ;
  if (!display->scene()) 
    return DTK_CALLBACK_ERROR ;

  //  Get the mouse pointer
  pointer = (dtkInValuator *) app->get("pointer", DTKINVALUATOR_TYPE);

  if(!pointer || pointer->number() != 2) 
    return DTK_CALLBACK_ERROR;

  //  Initialize the intersection detection.  Segment 0 will contain a line
  //  segment starting at the head position and extending 100 units in the 
  //  direction of the mouse pointer.  This will be in scene coordinates.
  intersect = new dpfIntersection();
  intersect->enableSegment(0);
  intersect->setRoot(display->scene());
  intersect->setSegmentLength(0, 100);
  
  //  Load the gnomon model
  char *path = (char *) pfGetFilePath();
  if(path)
    // copy path from static memory to the heap.
    path = dtk_strdup(path);

  pfFilePath(".:"DPF_DATA_DIR);

  pfGroup *dcs =
    new dpfDCS(gnomonPos, /* position */
		 0.03f, /* overall scale */
		 /* file       scale xyz     position xyz  rotation hpr */
		 "redSphere.pfb", DPF_IDENTITY,
		 NULL);

  if(path)
    {
      pfFilePath(path);
      free(path);
    }

  //  Disable intersection calculation for the gnomon model
  display->addGhostNode(dcs) ;
  display->scene()->addChild(dcs) ;

  //  Get the field of view
  float fov = display->masterScreen()->fov();

  //  Compute the distance to the viewing window, assuming the window is 2
  //  units wide (-1.0 to 1.0) with horizontal angle fov

  // Might want to tighten this up a bit later to be more forgiving of oter
  // window sizes and fov angles

  h = 1.0 / tan ( (M_PI / 180.0) * fov / 2.0);

  return DTK_CONTINUE ;
}


int mouseSelector::preFrame(void)
{

  float xy[2];
  pointer->read(xy); // poll the pointer

  //  Scale the y position by the aspect ratio
  int xSize, ySize;
  display->first()->first()->pfpwin()->getSize (&xSize, &ySize);
  xy[1] *= (float) ySize / (float) xSize;
  

  //  Set up segment 0 and do the intersection
  pfCoord* head = display->headPos();
  intersect->setSegmentPos(0, head->xyz);
  intersect->setSegmentDir(0, xy[0], h, xy[1]);
  intersect->doIntersection();

  //  If we got a hit, set the gnomon position to be the position of the
  //  intersection point
  if (intersect->segmentHit(0)) {
    pfVec3 intPoint;
    intersect->getXFormPoint(0, &intPoint);
    gnomonPos[0] = intPoint[0];
    gnomonPos[1] = intPoint[1];
    gnomonPos[2] = intPoint[2];
    
    //  Put the local intersection point in shared mem
    intersect->getLocalPoint(0, &intPoint);
    float point[3];
    point[0] = intPoint[0];
    point[1] = intPoint[1];
    point[2] = intPoint[2];
    shm->write(point, 3*sizeof(float));
  }
  else {
    gnomonPos[1] = -100;
    shm->write(gnomonPos, 3*sizeof(float));
  }
  
  return DTK_CONTINUE ;
}


/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new mouseSelector(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}


